/*exec.c****************************************************************
**version 0.2
*/

/*
** Code for executing execution trees.
*/

/*
** Stack Structure Rules:
**
** On entry to a new node (Exec_Next) (FP = FramePointer):
**
**(FramePointer is set such that ppop will restore the
** previous framepointer.  It thus points one space beyond
** this.)
**	(i) FP - 2 is empty for return value
**	(ii) FP - 1 is value of old FP
**	(iii) FP is address of node to execute,
**	(iv) FP + 1 is status of current node
**		(0 = entered from above,
**		 +ve depends on node details,
**		 -ve might eventually be error conditions ?)
**	(v) FramePointer + 2 and up are the values of any args
**		already evaluated, temp variables for this node,
**		and other miscelany.
**
** On return to higher node (Exit_Node):
**	(i) return value must already have been copied onto FP - 1
**	(ii) nothing above FP - 1 must be wanted
*/

#ifdef _DEBUGS_
#include <assert.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>

#define _EXEC_C_
#include "exec.h"

/* For printing verbose errors if deserialization fails */
#if 1
#define TRACEBACK(msg) printf(msg)
#else
#define TRACEBACK(msg) ((void)0)
#endif


static struct ENode *Deserialize_ENode(FILE *f);
static struct ENode *Build_Cnst(EINT32 value);

static void FTab_Unary(struct ENode *where,struct Tab *tab,FILE *f,int gap,struct Tab *prev);
static void FTab_First(struct ENode *where,struct Tab *tab,FILE *f,int gap,struct Tab *prev);
static void FTab_Middle(struct Tab *tab,FILE *f);
static void FTab_Last(struct Tab *tab,FILE *f);

/* Memory management functions
** These two are to be replaced by something more efficient later
*/

INLINE struct ENode *enode_malloc(void) { return malloc(sizeof(struct ENode)); }
INLINE void enode_free(void *ptr) { free(ptr); }

/* MACRO's for useful stack locations etc etc. */

/* Status of current node. */
#define STATUS ((FramePointer+1)->i)

/* integer/float and pointer returns resp. */
#define IRETVAL ((FramePointer-2)->i)
#define FRETVAL ((FramePointer-2)->f)
#define PRETVAL ((FramePointer-2)->p)

/* These will point to int/float and pointer returns from our
** children *if* (i) we allocated the return-space @ location per
** child *and* (ii) no local variables.
**
** (i) *WAS* always going to be true *but* there might be some point in
**     using only one location for several nodes
** (ii) OK so it isn't that good a MACRO (:-)
*/

#define IARG(x) ((FramePointer+2+x)->i)
#define FARG(x) ((FramePointer+2+x)->f)
#define PARG(x) ((FramePointer+2+x)->p)

/* push an int */

INLINE void ipush(EINT32 val) {
#ifdef _DEBUGS_
  assert((StackPointer - Stack) < (STACK_SIZE - 1));
#endif
  (StackPointer++)->i = val;
}

/* push a float */

INLINE void fpush(float val) {
#ifdef _DEBUGS_
  assert((StackPointer - Stack) < (STACK_SIZE - 1));
#endif
  (StackPointer++)->f = val;
}

/* push a pointer */

INLINE void ppush(void *val) {
#ifdef _DEBUGS_
  assert((StackPointer - Stack) < (STACK_SIZE - 1));
#endif
  (StackPointer++)->p = val;
}

/* pop an int */

INLINE EINT32 ipop() {
#ifdef _DEBUGS_
  assert((StackPointer - Stack) > 0);
#endif
  return (--StackPointer)->i;
}

/* pop a float */

INLINE float fpop() {
#ifdef _DEBUGS_
  assert((StackPointer - Stack) > 0);
#endif
  return (--StackPointer)->f;
}

/* pop a pointer */

INLINE void *ppop() {
#ifdef _DEBUGS_
  assert((StackPointer - Stack) > 0);
#endif
  return (--StackPointer)->p;
}

/*
** Sets up stack for entry to a lower node:
**
**		(i) Push 0 (for return value slot)
**			Now left to be done by routines themselves.
**		(ii) Push FramePointer
**		(iii) Set FramePointer = StackPointer
**		(iv) Push pointer to new node;
**		(v) Push 0 (status of any node on entry)
*/

INLINE void Enter_Node(struct ENode *which) {
/*  ipush(0); Now do in individual routines, to give choice about
**            return values from multiple children. */

  /* In here I suggest some graceful failure if the program is running
  ** short of stack, and penalise the probe to discourage them from
  ** using such verbose programs. - Toby
  */

  ppush(FramePointer);
  FramePointer = StackPointer;
  ppush(which);
  ipush(0);
}

/*
** Performs Tidy-up when leaving a node in the upwards direction.
** Discards any items above the FramePointer on the stack and restores
** the FramePointer to what it was for the previous node.
**
** This leaves the return value as the top item on the stack.
*/

INLINE void Exit_Node() {
  StackPointer = FramePointer;  FramePointer = ppop();
}

/* Executes one step of the current node */

int Exec_Next() {
  struct ENode *where;

  where = FramePointer->p;
  return where->functions->exe(where);
}

/* Executes an entire tree (this is for testing I suppose.
** First node must know to return 0 when tree is complete. */

int Exec_Whole(struct ENode *top) {
  ipush(0);	/* for final return value */
  Enter_Node(top);
  
  while(Exec_Next());
  return ipop();
}  

/*
** Stack sketch:
**
** ...
** ^
** local values of previous node
** v
** ^
** returns from older brothers/sisters (what do you mean you can't tell the
**                                      sex of a node !)
** v
** our return value
** old FramePointer
** node		<- FramePointer points here
** our status
** ^
** our local variables
** v
** ^
** returns from any children already exavluated
** v
** 		<- StackPointer points here
** ...
*/

/***************************************************************************/
/************************** ENodes Follow **********************************/
/***************************************************************************/

/******************************* TOP *******************************
** Place marker for top of a tree.
** The only one that returns 0 to signify completion.
** (irrelevane to Alife ?)
*/

struct TOP_private
{
  struct ENode *body;
};

#define TOP_BODY (((struct TOP_private *)(where->private))->body)

int TOP_exe(struct ENode *where) {
  if(!STATUS++)
  {
    ipush(0);
    Enter_Node(TOP_BODY);
    return 1;
  }
  else
  {
    IRETVAL = IARG(0);
    Exit_Node();
    return 0;
  }
}

void TOP_serialize(struct ENode *where,FILE *f,struct Tab *prev_tab)
{
  struct Tab tab;

  fprintf(f,"TOP");

  FTab_Unary(where,&tab,f,3,prev_tab);
  TOP_BODY->functions->serialize(TOP_BODY,f,&tab);
}

struct ENode *TOP_deserialize(struct ENode *where,FILE *f)
{
  where->private=malloc(sizeof(struct TOP_private));
  if(where->private)
  {
    if(NULL!=(TOP_BODY=Deserialize_ENode(f)))
    {
      return where;
    }
    else
    {
      TRACEBACK("...TOP\n");
    }
  }
  else
  {
    TRACEBACK("TOP allocation failed\n");
  }

  /* Reaching here means something failed */

  if(where->private)
  {
    if(where->private) free(where->private);
  }
  enode_free(where);
  return NULL;    
}

int TOP_volume(struct ENode *where,int paint)
{
  if(paint) paint=paint-1;
  
  if(paint) paint=TOP_BODY->functions->volume(TOP_BODY,paint);
  return paint;                                   
}

void TOP_delete(struct ENode *where)
{
  TOP_BODY->functions->delete(TOP_BODY);
  free(where->private);
  enode_free(where);
}

static const struct ENodeFunctions TOP =
{
  TOP_exe,
  TOP_serialize, TOP_deserialize,
  TOP_volume,
  TOP_delete,

  "TOP", NULL
};



/******************************* PLUS *******************************
** Adds its (int) children *to_whatever_is_already_in* it's return value.
** (e.g lets father decide whether we're summing from 0 or accumulating.
*/

struct PLUS_private
{
  struct ENode *left,*right;
};

#define PLUS_LEFT (((struct PLUS_private *)(where->private))->left)
#define PLUS_RIGHT (((struct PLUS_private *)(where->private))->right)

int PLUS_exe(struct ENode *where) {

  switch(STATUS++)
  {
    case 0:
      ipush(0);
      Enter_Node(PLUS_LEFT);
      break;

    case 1:
      IRETVAL+=ipop();
      ipush(0);
      Enter_Node(PLUS_RIGHT);
      break;

    case 2:
      IRETVAL+=ipop();
      Exit_Node();
      break;
  }

  return 1;
}

/* Store the PLUS ENode and both its child nodes in a file.
** This function does no error checking for disk full etc,
** it is the reposonibility of the function that opened
** the file to check ferror after serialization is complete
*/

void PLUS_serialize(struct ENode *where,FILE *f,struct Tab *prev_tab)
{
  struct Tab tab;

  /* Always print our identifier first, this will be used in
  ** deserialization to determine which type of ENode to create
  */
  fprintf(f,"PLUS");

  /* The variable tab keeps track of the tabstop information required
  ** for drawing tabs into our child nodes. FTab_First initialises it,
  ** links it to the tabstop information to our parent node, and draws
  ** the first tab.
  ** Immediately after calling a tab function we call the serialize
  ** function for a child node.
  */
  FTab_First(where,&tab,f,4,prev_tab);
  PLUS_LEFT->functions->serialize(PLUS_LEFT,f,&tab);

  /* This enode has more than one daughter Enode; before serializing
  ** each one we must call a tab function. Here FTab_Last is called
  ** since this is the last child of this node. If we had three
  ** or more child nodes then we would call FTab_Middle before 
  ** serializing nodes other than the first and the last (for example
  ** see AIF). If we had only one child then the call to FTab_First
  ** would be replaced with FTab_Unary (for example see TOP).
  */
  FTab_Last(&tab,f);
  PLUS_RIGHT->functions->serialize(PLUS_RIGHT,f,&tab);
}

struct ENode *PLUS_deserialize(struct ENode *where,FILE *f)
{
  /* This function is called by Deserialize_ENode after the
  ** ENode name has been read from the file and in this case
  ** matched with the string "PLUS". This function is passed
  ** an ENode structure in its where paramater which is
  ** initialized except for the member private.
  ** This function must perform any initialization required
  ** for this ENode, which includes creating any child
  ** nodes. It should return where if everything succeeds, or
  ** NULL if it failed.
  */
  
  where->private=malloc(sizeof(struct PLUS_private));
  if(where->private)
  {
    if(NULL!=(PLUS_LEFT=Deserialize_ENode(f)))
    {
      if(NULL!=(PLUS_RIGHT=Deserialize_ENode(f)))
      {
        /* Everything succeeded */
        return where;
      }
      else
      {
        TRACEBACK("...PLUS, second operand\n");
      }
    }
    else
    {
      TRACEBACK("...PLUS, first operand\n");
    }
  }
  else
  {
    TRACEBACK("PLUS memory allocation failed\n");
  }

  /* Reaching here means something went wrong */

  if(where->private)
  {
    if(PLUS_LEFT)
    {
      PLUS_LEFT->functions->delete(PLUS_LEFT);
    }
    free(where->private);
  }
  enode_free(where);
  return NULL;
}

int PLUS_volume(struct ENode *where,int paint)
{
  /* The volume function is used to quantify the size of the
  ** subtree below this ENode, whether it is larger or smaller
  ** than a given size. This is defined recursively using a
  ** painting analogy; this ENode is given an amount of paint
  ** with which to cover its subtree, and it should return the
  ** amount of paint remaining after the subtree has been covered
  */ 

  /* If there is any paint left, use up one unit for this node */
  if(paint) paint=paint-1;

  /* If there is still paint left, use it to paint the left subtree */
  if(paint) paint=PLUS_LEFT->functions->volume(PLUS_LEFT,paint);
  
  /* If there is still paint left, use it to paint the other subtree */
  if(paint) paint=PLUS_RIGHT->functions->volume(PLUS_RIGHT,paint);

  /* Return the amount remaining */
  return paint;                                   
}

void PLUS_delete(struct ENode *where)
{
  PLUS_LEFT->functions->delete(PLUS_LEFT);
  PLUS_RIGHT->functions->delete(PLUS_RIGHT);
  free(where->private);
  enode_free(where);
}

static const struct ENodeFunctions PLUS =
{
  PLUS_exe,
  PLUS_serialize, PLUS_deserialize,
  PLUS_volume,
  PLUS_delete,

  "PLUS", &TOP
};


/******************************* AIF *******************************
** Arithmetic if; Evaluates one expression, the evaluates and returns
** one of its other two subtreees depending on whether the first
** returned zero or non-zero
*/

struct AIF_private
{
  struct ENode *test,*iftrue,*iffalse;
};

#define AIF_TEST (((struct AIF_private *)(where->private))->test)
#define AIF_IFTRUE (((struct AIF_private *)(where->private))->iftrue)
#define AIF_IFFALSE (((struct AIF_private *)(where->private))->iffalse)

int AIF_exe(struct ENode *where) {

  switch(STATUS++)
  {
    case 0:
      ipush(0);
      Enter_Node(AIF_TEST);
      break;

    case 1:
      if(ipop())
      {
        ipush(0);
        Enter_Node(AIF_IFTRUE);
      }
      else
      {
        ipush(0);
        Enter_Node(AIF_IFFALSE);
      }
      break;

    case 2:
      IRETVAL=ipop();
      Exit_Node();
      break;
  }

  return 1;
}

void AIF_serialize(struct ENode *where,FILE *f,struct Tab *prev_tab)
{
  struct Tab tab;
  
  fprintf(f,"AIF");
  
  FTab_First(where,&tab,f,3,prev_tab);
  AIF_TEST->functions->serialize(AIF_TEST,f,&tab);
  FTab_Middle(&tab,f);
  AIF_IFTRUE->functions->serialize(AIF_IFTRUE,f,&tab);
  FTab_Last(&tab,f);
  AIF_IFFALSE->functions->serialize(AIF_IFFALSE,f,&tab);
}

struct ENode *AIF_deserialize(struct ENode *where,FILE *f)
{
  where->private=malloc(sizeof(struct AIF_private));
  if(where->private)
  {
    if(NULL!=(AIF_TEST=Deserialize_ENode(f)))
    {
      if(NULL!=(AIF_IFTRUE=Deserialize_ENode(f)))
      {
        if(NULL!=(AIF_IFFALSE=Deserialize_ENode(f)))
        {
          return where;
        }
        else
        {
          TRACEBACK("...PLUS, operand if false\n");
        }
      }
      else
      {
        TRACEBACK("...PLUS, operand if true\n");
      }
    }
    else
    {
      TRACEBACK("...AIF, test operand\n");
    }
  }
  else
  {
    TRACEBACK("AIF, allocation failed\n");
  }

  /* Reaching here means somethign failed */

  if(where->private)
  {
    if(AIF_TEST)
    {
      if(AIF_IFTRUE)
      {
        AIF_IFTRUE->functions->delete(AIF_IFTRUE);
      }
      AIF_TEST->functions->delete(AIF_TEST);
    }
    free(where->private);
  }
  enode_free(where);
  return NULL;
}

int AIF_volume(struct ENode *where,int paint)
{
  if(paint) paint=paint-1;
  
  if(paint) paint=AIF_TEST->functions->volume(AIF_TEST,paint);
  
  if(paint) paint=AIF_IFTRUE->functions->volume(AIF_IFTRUE,paint);
  
  if(paint) paint=AIF_IFFALSE->functions->volume(AIF_IFFALSE,paint);
  
  return paint;                                   
}

void AIF_delete(struct ENode *where)
{
  AIF_TEST->functions->delete(AIF_TEST);
  AIF_IFTRUE->functions->delete(AIF_IFTRUE);
  AIF_IFFALSE->functions->delete(AIF_IFFALSE);
  free(where->private);
  enode_free(where);
}

static const struct ENodeFunctions AIF =
{
  AIF_exe,
  AIF_serialize, AIF_deserialize,
  AIF_volume,
  AIF_delete,

  "AIF", &PLUS
};


/******************************* CNST *******************************
** Returns a constant.
*/

struct CNST_private
{
  EINT32 value;
};

#define CNST_VALUE (((struct CNST_private *)(where->private))->value)

int CNST_exe(struct ENode *where) {
  IRETVAL = CNST_VALUE;
  Exit_Node();
  return 1;
}

void CNST_serialize(struct ENode *where,FILE *f,struct Tab *prev_tab)
{
  fprintf(f,"%ld",(long)CNST_VALUE);
}

int CNST_volume(struct ENode *where,int paint)
{
  if(paint) paint=paint-1;
  return paint;
}

void CNST_delete(struct ENode *where)
{
  free(where->private);
  enode_free(where);
}

static struct ENodeFunctions CNST =
{
  CNST_exe,
  CNST_serialize, NULL,
  CNST_volume,
  CNST_delete,

  "CNST", &AIF
};

static const struct ENodeFunctions *FirstENodeFunctions=&CNST;

static struct ENode *Deserialize_ENode(FILE *f)
{
  /* Slow, but easily extensible scanner. This function will be
  ** called by deserialization functions to decode an ENode from a
  ** file, and in turn it recursively calls the deserialization
  ** functions for child enodes.
  ** This function uses a linear search through the linked list of
  ** ENodeFunction blocks. When the set of ENode types is stable this
  ** can be replaced with a faster search if speed is a problem.
  */

  static const char whitespace_table[]="\r\n\t|\\_ ";
  static const char numbers_table[]="-0123456789";

  /* CAUTION: static locals in a recursive function! */
  static int only_digits_in_this_name_so_far;
  static char name[ENODE_NAME_LEN+1]; 
  static int i,c;
  static const struct ENodeFunctions *functions;
  static struct ENode *enode;

  only_digits_in_this_name_so_far=1;
  
  /* Scan past whitespace */
  for(;;)
  {
    c=fgetc(f);
    if(EOF==c) return NULL;
    if(!strchr(whitespace_table,c)) break;
  }

  /* Scan up to next whitespace */
  i=0;
  for(;;)
  {
    if(i>=ENODE_NAME_LEN)
    {
      TRACEBACK("Name too long\n");
      return NULL;
    }

    if(only_digits_in_this_name_so_far && (!strchr(numbers_table,c)) )
      only_digits_in_this_name_so_far=0;

    name[i]=c;
    name[++i]='\0';

    c=fgetc(f);
    if(EOF==c) break;
    if(strchr(whitespace_table,c)) break;
  }

  if(only_digits_in_this_name_so_far)
  {
    /* Ugly kludge: Treat CNST ENodes differently because they */
    /* are serialized just as their digits, NOT including the ENode name. */
    return Build_Cnst(atol(name));
  }
  else
  {
    /* Now search for ENode types matching name */
    functions=FirstENodeFunctions;
    while(functions)
    {
      if( (!strcmp(name,functions->name)) )
      {
        /* A match! 
        ** From here down we do not need any of our local variables,
        ** but most compilers will not pick up on this and will
        ** keep them on the stack. They have been made static
        ** since stack space is precious in this recursive function
        */
        enode=enode_malloc();
        if(enode)
        {
          enode->functions=functions;
          enode->private=NULL;
          enode=enode->functions->deserialize(enode,f);
        }
        else
        {
          TRACEBACK("ENode allocation failed\n");
        }
        return enode;
      }
      functions=functions->next;
    }

    /* No match */
    TRACEBACK("No match\n");
    return NULL;
  }
}

static struct ENode *Build_Cnst(EINT32 value)
{
  struct ENode *enode;
  enode=enode_malloc();
  if(enode)
  {
    enode->functions=&CNST;
    enode->private=malloc(sizeof(struct PLUS_private));
    if(enode->private)
    {
        ((struct CNST_private *)(enode->private))->value=value;
    }
    else
    {
      TRACEBACK("CNST allocation failed\n");
      enode_free(enode);
      enode=NULL;
    }
  }
  else
  {
    TRACEBACK("ENode allocation failed\n");
  }
  return enode;
}

/* When serializing nodes with only one child node:
**      FTab_Unary(...);
**      CHILD->functions->serialize(CHILD,...);
**
** When serializing nodes with only two child nodes:
**      FTab_First(...);
**      CHILD->functions->serialize(CHILD,...);
**      FTab_Last(...);
**      CHILD->functions->serialize(CHILD,...);
**
** When serializing nodes with more than two child nodes:
**      FTab_First(...);
**      CHILD->functions->serialize(CHILD,...);
**      FTab_Middle(...);
**      CHILD->functions->serialize(CHILD,...);
**      .....
**      FTab_Middle(...);
**      CHILD->functions->serialize(CHILD,...);
**      FTab_Last(...);
**      CHILD->functions->serialize(CHILD,...);
*/

static const char Tree_Section[5][4]=
{ "___",
  " |_",
  " | ",
  " \\_",
  "   " };

static void FTab_Unary(struct ENode *where,struct Tab *tab,FILE *f,int gap,struct Tab *prev)
{
  FTab_First(where,tab,f,gap,prev);
  tab->state=0;
}

static void FTab_First(struct ENode *where,struct Tab *tab,FILE *f,int gap,struct Tab *prev)
{
  /* The paint constant specifies the largest subtree draw on one line. */
  const int paint=7;
  
  if(where->functions->volume(where,paint))
  {
    /* This is close to the end of the tree, draw remainder on one line. */
    tab->enabled=0;
    fprintf(f," ");
  }
  else
  {
    /* Lots of subtree left, draw this part at least on multiple lines */
    tab->enabled=1;
    tab->prev=prev;
    tab->gap=gap;
    tab->state=1;
    fprintf(f,Tree_Section[0]);
  }
}

static void FTab_Internal(struct Tab *tab,FILE *f)
{
  int i;

  if(tab->prev) FTab_Internal(tab->prev,f);
  else          fprintf(f,"\n");

  for(i=0;i<tab->gap;++i) fputc(' ',f);

  if(tab->state) fprintf(f,Tree_Section[2]);
  else           fprintf(f,Tree_Section[4]);
}

static void FTab_Middle(struct Tab *tab,FILE *f)
{
  int i;
  
  if(tab->enabled)
  {  
    if(tab->prev) FTab_Internal(tab->prev,f);

    for(i=0;i<tab->gap;++i) fputc(' ',f);
    fprintf(f,Tree_Section[1]);
  }
  else
  {
    fprintf(f," ");
  }    
}

static void FTab_Last(struct Tab *tab,FILE *f)
{
  int i;

  if(tab->enabled)
  {  
    if(tab->prev) FTab_Internal(tab->prev,f);
    tab->state=0;

    for(i=0;i<tab->gap;++i) fputc(' ',f);
    fprintf(f,Tree_Section[3]);
  }
  else
  {
    fprintf(f," ");
  }
}


/************************* Test Routine **************************************/

void Walk_ENodeFunctions(void)
{
  /* Walk the linked list of ENodeFunctions, printing the name of each type
  ** of ENode. If it prints garbage you know the linkage between
  ** ENodeFunctions is bad.
  */

  const struct ENodeFunctions *functions;
  char *type;
  
  printf("This version decodes the following ENodes:\n");
  functions=FirstENodeFunctions;
  while(functions)
  {
    printf("%s, ",functions->name);
    functions=functions->next;
  }
  printf("\n");
}

int main()
{
  struct ENode *enode;
  FILE *f;

  Walk_ENodeFunctions();
  
  f=fopen("execin.gpt","r");
  if(NULL!=f)
  {
    enode=Deserialize_ENode(f);
    fclose(f);
    if(enode)
    {
      printf("%i\n", Exec_Whole(enode));

      f=fopen("execout.gpt","w");
      if(NULL!=f)
      {
        enode->functions->serialize(enode,f,NULL);
        fclose(f);
      }

      enode->functions->delete(enode);
    }
  }
  
  return 0;
}

